/* Diffie-Hellman key exchange algorithm */
/* both client and server call this function : */
/* int diffie_hellman_handshake(int fd, long *input_seed, long *output_seed, int is_server) */
/* the resulting input and output seeds are privately negotiated session keys */

/* Diffie-Hellman public key algorithm :
   Patent expired in 1997, this algorithm is now
   in the public domain.  This is similar to RSA,
   but based on the mathematical difficulty of
   computing discrete logarithms (rather than
   factoring).  Diffie-Hellman was the first
   public key algorithm ever invented, and it
   is still widely used.
*/

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

/* we require GNU's MP library for arbitrary precision math */

#include <gmp.h>

/* string representations of integers are in hexadecimal */
#define DHBASE 16

/* number of repetitions to make sure a number is actually prime */
/* see gmp documentation for more info, but 32 should be plenty */
#define DHREPS 32

extern void srand48() ;
extern long lrand48() ;

void sure_ascii_write(long *output_seed,int fd,char *s) ;  /* prototype */

void dhcrypt_choose_large_prime(int num_bits, char *result)
{
  mpz_t prime, temp, one ;
  struct timeval now ;
  char *seed ;
  int i, nibble, done ;
  unsigned long rs ;

  mpz_init(prime) ;
  mpz_init(temp) ;
  mpz_init(one) ;

  mpz_set_str(one, "1", DHBASE) ;

  seed=(char *)malloc((num_bits/4)+2) ;
  gettimeofday(&now,NULL) ;
  rs=((unsigned long) getpid())*now.tv_usec + now.tv_sec ;
  rs^=(unsigned long) &rs ;
  srand48(rs) ;
  for(i=0;i<num_bits/4;i++)
    {
      rs=rs>>1 ;
      nibble=(lrand48()+rs)%16 ;
      if(nibble<0)
	nibble= (nibble * -1)%16 ;
      if(nibble<10)
	seed[i]='0'+nibble ;
      else
	seed[i]='A'+nibble-10 ;
      seed[i+1]=0 ;
    }
  mpz_set_str(prime, seed, DHBASE) ;

  /* fprintf(stderr,"dhcrypt_choose_large_prime : starting with [%s]\n",seed) ; */
      
  for(done=0;!done;)
    {
      if(mpz_probab_prime_p(prime, DHREPS))
	done=1 ;
      else
	{
	  mpz_add(temp,prime,one) ;
	  mpz_set(prime,temp) ;
	}
    }

  mpz_get_str(result, DHBASE, prime) ;
  /* fprintf(stderr,"dhcrypt_choose_large_prime : stopping on [%s]\n",result) ; */

  mpz_clear(prime) ;
  mpz_clear(temp) ;
  mpz_clear(one) ;
}

void dhcrypt_compute(char *base_p,char *exponent_p,char *modulus_p, char *result_p)
{
  mpz_t base, exponent, modulus, result ;

  mpz_init(base) ;
  mpz_init(exponent) ;
  mpz_init(modulus) ;
  mpz_init(result) ;

  mpz_set_str(base, base_p, DHBASE) ;
  mpz_set_str(exponent, exponent_p, DHBASE) ;
  mpz_set_str(modulus, modulus_p, DHBASE) ;

  mpz_powm(result,base,exponent,modulus) ;

  mpz_get_str(result_p, DHBASE, result) ;

  mpz_clear(base) ;
  mpz_clear(exponent) ;
  mpz_clear(modulus) ;
  mpz_clear(result) ;
}

/* returns 0 on success */
int diffie_hellman_handshake(int fd, long *input_seed, long *output_seed, int is_server)
{
  int error_result=1 ;
  char *base=NULL ;
  char *modulus[2]={NULL,} ;
  char *his_public_key[2]={NULL,} ;
  char *my_secret_key[2] ;
  char *my_public_key[2] ;
  char *negotiated_session_key[2] ;
  char temps[2048] ;
  char tempbuf[2048] ;
  char *buf, *p, *lparen ;
  int i, j, x, done, num_rparens ;
  struct timeval timeout,start,now ;
  fd_set readfds ;
  char *input_seed_string, *output_seed_string ;

  for(i=0;i<2;i++)
    {
      my_secret_key[i]=malloc(128) ;
      my_public_key[i]=malloc(128) ;
      negotiated_session_key[i]=malloc(128) ;

      sprintf(my_public_key[i],"0") ;
      sprintf(negotiated_session_key[i],"0") ;

      dhcrypt_choose_large_prime(128,my_secret_key[i]) ;
    }

  gettimeofday(&start,NULL) ;
  if(is_server)
    {
      base="2" ;
      sprintf(temps,"(%s)",base) ;
      sure_ascii_write(NULL,fd,temps) ;
      for(i=0;i<2;i++)
        {
          modulus[i]=malloc(128) ;
          dhcrypt_choose_large_prime(128,modulus[i]) ;
          sprintf(temps,"(%s)",modulus[i]) ;
          sure_ascii_write(NULL,fd,temps) ;
        }
      for(i=0;i<2;i++)
        {
          dhcrypt_compute(base,my_secret_key[i],modulus[i],my_public_key[i]) ;
          fprintf(stderr,"my pubkey[%d]=[%s]\n",i,my_public_key[i]) ;
          sprintf(temps,"(%s)",my_public_key[i]) ;
          sure_ascii_write(NULL,fd,temps) ;
        }

      num_rparens=3 ;
    }
  else
    num_rparens=0 ;
  buf=tempbuf ;
  for(done=0;!done;)
    {
      gettimeofday(&now,NULL) ;

      timeout.tv_sec=32-(now.tv_sec-start.tv_sec) ;
      if(timeout.tv_sec<=0)
	done=1 ;
      else
	{
	  timeout.tv_usec=0 ;

	  FD_ZERO(&readfds) ;
	  FD_SET(fd,&readfds) ;
	  select(fd+1,&readfds,NULL,NULL,&timeout) ;
	  if(FD_ISSET(fd,&readfds))
	    {
	      int r=read(fd,buf,1024) ;
	      if(r==0)
		done=1 ;  /* connection dropped */
	      if(r>0)
		{
		  for(p=buf;p<buf+r;p++)
		    {
		      if(p[0]=='(')
			lparen=p ;
		      if(p[0]==')')
			{
			  p[0]=0 ;
			  switch(num_rparens++)
			    {
			    case 0 :
			      base=lparen+1 ;
			      break ;;;
			    case 1 :
			      modulus[0]=lparen+1 ;
			      break ;;;
			    case 2 :
			      modulus[1]=lparen+1 ;
			      break ;;;
			    case 3 :
			      his_public_key[0]=lparen+1 ;
			      break ;;;
			    case 4 :
			      his_public_key[1]=lparen+1 ;
			      done=1 ;
			      break ;;;
			    }
			}
		    }
		  buf+=r ;
		}
	    }
	}
    }

  /*
  fprintf(stderr,"base=[%s]\n",base) ;
  fprintf(stderr,"modulus[0]=[%s]\n",modulus[0]) ;
  fprintf(stderr,"modulus[1]=[%s]\n",modulus[1]) ;
  fprintf(stderr,"his_public_key[0]=[%s]\n",his_public_key[0]) ;
  fprintf(stderr,"his_public_key[1]=[%s]\n",his_public_key[1]) ;
  fprintf(stderr,"my_secret_key[0]=[%s]\n",my_secret_key[0]) ;
  fprintf(stderr,"my_secret_key[1]=[%s]\n",my_secret_key[1]) ;
  */

  if(num_rparens>=5)
    {
      error_result=0 ;

      for(i=0;i<2;i++)
	{
          if(!is_server)
            {
              dhcrypt_compute(base,my_secret_key[i],modulus[i],my_public_key[i]) ;
              /* fprintf(stderr,"my pubkey[%d]=[%s]\n",i,my_public_key[i]) ; */
              sprintf(temps,"(%s)",my_public_key[i]) ;
              sure_ascii_write(NULL,fd,temps) ;
            }
	  dhcrypt_compute(his_public_key[i],my_secret_key[i],modulus[i],negotiated_session_key[i]) ;
	  /* fprintf(stderr,"nsk[%d]=[%s]\n",i,negotiated_session_key[i]) ; */
	}

      if(is_server)
	x=0 ;
      else
	x=1 ;
      input_seed_string=negotiated_session_key[x] ;
      output_seed_string=negotiated_session_key[1-x] ;
      /* reduce 128 bit session key down to 32 bit transport encryption key */
      input_seed_string+=(strlen(input_seed_string)-2*sizeof(long)) ;
      output_seed_string+=(strlen(output_seed_string)-2*sizeof(long)) ;
      if((1!=sscanf(output_seed_string,"%lx",output_seed)) ||
	 (1!=sscanf(input_seed_string,"%lx",input_seed)))
	{
	  perror("diffie hellman handshake: sscanf nsk failed") ;
	  exit(-1) ;
	}
      /*
      fprintf(stderr,"input seed=0x%lX\n",*input_seed) ;
      fprintf(stderr,"output seed=0x%lX\n",*output_seed) ;
      */
    }

  for(i=0;i<2;i++)
    {
      free(my_secret_key[i]) ;
      free(my_public_key[i]) ;
      free(negotiated_session_key[i]) ;
    }

  return(error_result) ;
}
